#!/usr/bin/env python
#!/usr/bin/python
#
# c3.py
# program to crossmatch objects from two different catalogues
#
# Author: G. Riccio
# Creation Date: March 1, 2016
# Release: 1.0
# Release Date: July 23, 2016
#
# Copyright 2016, Giuseppe Riccio, Massimo Brescia
# This file is part of C3.

    # C3 is free software: you can redistribute it and/or modify
    # it under the terms of the GNU General Public License as published by
    # the Free Software Foundation, either version 3 of the License, or
    # (at your option) any later version.

    # C3 is distributed in the hope that it will be useful,
    # but WITHOUT ANY WARRANTY; without even the implied warranty of
    # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    # GNU General Public License for more details.

    # You should have received a copy of the GNU General Public License
    # along with C3.  If not, see <http://www.gnu.org/licenses/>.
#
### WINDOWS INSTRUCTIONS
#
# This program has been produced with python 2.7.8 (64bit)
# It has been tested on Windows 7 and 10, by running it in a system
# shell window.
#
# IMPORTANT: be sure to have previously installed python, together with the following external packages:
#
# 	numpy;
#	pyfits;
#
### LINUX UBUNTU INSTRUCTIONS
#
# This program has been tested on Linux Ubuntu 14.04 with python 2.7.6 (64-bit)
#
#################
# SYNOPSIS
#
# Given two catalogues, this multithread code makes the crossmatch based on the mutual position.
#
###################
# USAGE
#
# prompt>> python c3.py [-c (or --config) configfile]
#
# parameters:
#	-c, --config: [optional], set the configuration file name 
#   
# EXAMPLE OF CONFIGURATION FILE
#
#[I/O Files]
#
#Input catalogue 1:	./file1.csv
#Format catalogue 1:	csv				;csv, fits, votable or ascii
#
#Input catalogue 2:	./file2.csv
#Format catalogue 2:	csv				;csv, fits, votable or ascii
#Output:			./output/output.csv
#Output format:		csv				;csv, fits, votable or ascii
#Log file:		./output/out.log
#Stilts directory:	./libs
#working directory:	./tmp				;temporary directory, removed after completed
#
#[Match Criteria]
#algorithm:		sky				;sky, exact value, row-by-row
#
#[Exact Value parameters]
#;required for algorithm: "exact value"			
#Input 1 column:		ID				;match column catalog 1
#Input 2 column:		1				;match column catalog 2
#
#[Row by Row parameters]
#;no specific parameters are required for algorithm: "row-by-row"			
#
#[Sky parameters]
#;required for algorithm: "sky"
#area shape:		ellipse				;ellipse or rectangle
#size type:		fixed				;parametric or fixed
#matching area first dimension:		10		;arcsec in case of fixed type - column name/number in case of parametric type 
#matching area second dimension:		10		;arcsec in case of fixed type - column name/number in case of parametric type 
#parametric factor:			1		;multiplicative factor for dimension columns - required for parametric type
#pa column/value:	45				;degrees in case of fixed type - column name/number in case of parametric type -missing or default = 0 deg
#pa settings:		clock, 0			;orientation (clock, counter), shift (degrees) - missing or default = clock,0)
#Catalogue 2 minimum partition cell size:125		;arcsec, (optional, default=100 arcsecs)
#
#[Catalogue 1 Properties]
#coordinate system:	galactic			;galactic, icrs, fk4, fk5
#coordinate units:	deg				;degrees (or deg), radians (or rad), sexagesimal (or sex)
#glon/ra column:		GLON				;column number or name - required for error type fwhm
#glat/dec column:	GLAT				;column number or name - required for error type fwhm
#designation column:	ID1				;column number or name - -1 for none
#
#
#[Catalogue 2 Properties]
#coordinate system:	icrs				;galactic, icrs, fk4, fk5
#coordinate units:	deg				;degrees (or deg), radians (or rad), sexagesimal (or sex)
#glon/ra column:		RA				;column number or name
#glat/dec column:	DEC				;column number or name
#designation column:	ID2				;column number or name - -1 for none - available only for area shape ellipse
#
#
#[Threads Properties]
#thread limit: 		4				;maximum number of simultaneous threads (it depends on your machine)
#
#[Output Rows]
#Match selection:	all				;all or best
#Join type:		1 and 2				;1 and 2, 1 or 2, all from 1, all from 2, 1 not 2, 2 not 1, 1 xor 2
#
##########################################################################################################################

# import needed libraries

import 	sys
import 	argparse
from    time			import time, localtime, asctime, sleep
#from 	string			import split

from    libs.utility	 	import findColMinMax, writeLogFile, getMinFWHM, writeTime, setSeparator,createConfig, errorExit, removeTmp
from    math	    		import floor, ceil
from    os	      		import path, sep, makedirs,system
from 	libs.threadSky		import skyRun
import  csv
#from    libs.resultsAnalysis 	import resultClumpAnalysis, resultSourcesAnalysis,countSources
import  shutil
from    libs.threadExact	import exactRun
from    libs.ioUtils	 	import *
from	libs.check		import checkFile, checkType, checkCatSize
from	libs.match		import runRowByRow, createOutput
import 	numpy as np
from 	libs.prePhase		import newRowsCat1,createSubDiv, splitRows, preExact


start=time()

######################## CONFIGURATION ###########################

def main(configfile, maindir):
	



	#check args
	if configfile==None:
		configfile="config.ini"
	
	#setlogfile
	print("\nC3 Starting Time: "+asctime(localtime(start)))
	logfile=getLogfile(configfile)
	
	writeLogFile(logfile, "C3 Starting Time: "+asctime(localtime(start))+"\n\n", "w")

	sys.stdout.write("\n\nInitializing...")
	sys.stdout.flush()	

	#get input files and directories for stilts and tmp; set output file	
	file1, file2, outputFile, format1, format2, outformat=getIO(configfile, logfile)

	ftype1=checkFile(file1, format1,logfile)
	ftype2=checkFile(file2, format2,logfile)
	fout=checkType(outformat, logfile)
	stiltsDir, tmp_path=getDirs(configfile, logfile)
	
	outputFile=setSeparator(outputFile)
	file1=setSeparator(file1)
	file2=setSeparator(file2)

	# set output dir & output files
	outdir, outfile=setOutDir(outputFile)




	#get Match Criteria
	criteria=readCriteria(configfile, logfile)

	file1tmp, header1, rows1=readFile(file1, ftype1, logfile, stiltsDir, tmp_path)
	file2tmp, header2, rows2=readFile(file2, ftype2, logfile, stiltsDir, tmp_path)	


	print(" DONE\n")

	writeLogFile(logfile, "********** I/O Data **********\n\n", "a")	
	writeLogFile(logfile, "Input catalogue 1: "+file1+"\n", "a")
	writeLogFile(logfile, "Input catalogue 2: "+file2+"\n\n", "a")

	writeLogFile(logfile, "Crossmatched catalogue: "+outdir+outfile+"\n", "a")	

	print("\nInput catalogue 1: "+file1)
	print("Input catalogue 2: "+file2+"\n")
	print("Crossmatched catalogue: "+outdir+outfile)

	checkCatSize(rows1, rows2, logfile, tmp_path)


	if criteria=="exact value" or criteria=="sky":
		n_threads=getNThreads(configfile, logfile, tmp_path)
		print("\nNumber of threads: " + str(n_threads)+"\n")
		writeLogFile(logfile, "\n\n************ THREADS INFO ************\n\n", "a")	
		writeLogFile(logfile, "Number of threads: " + str(n_threads)+"\n", "a")

	print("\nAlgorithm: "+criteria+"\n")
	writeLogFile(logfile, "\n\n************ MATCHING INFO ************\n\n", "a")
	writeLogFile(logfile, "Algorithm: " + criteria +"\n\n", "a")

	if criteria=="row-by-row":
		sys.stdout.write("Catalogues Matching Started...")
		sys.stdout.flush()
		t_row=time()
		writeTime(t_row, logfile)
		writeLogFile(logfile, "Crossmatching started\n", "a")	
		if len(rows1)==len(rows2):
			runRowByRow(file1tmp, rows1, file2tmp, rows2, header1, header2, outputFile)
			n_matches=len(rows1)
			t_final_row=time()-t_row
			print(" DONE (duration: %4.2fs)\n" % (t_final_row))
			writeLogFile(logfile, "\t\tCrossmatching phase duration: %4.2fs" % (t_final_row)+"\n\n", "a")
			shutil.rmtree(tmp_path)
		else:
			print("\nError! Input files dimensions mismatch. Please check and retry.")
			print("File "+file1+" --> "+str(len(rows1))+" records")
			print("File "+file2+" --> "+str(len(rows2))+" records\n")
			writeLogFile(logfile, "Error! Input files dimensions mismatch\n", "a")
			errorExit(tmp_path)

			
	elif criteria=="exact value":
		matchcol1, matchcol2=readMatchCol(header1, header2, configfile, logfile,tmp_path)
		selection, join=readSelection(configfile, logfile)
		print("Match Selection: "+selection+ " - Join Type: "+join)
		writeLogFile(logfile, "Match Selection: " + selection +"\n", "a")
		writeLogFile(logfile, "Join Type: " + join +"\n", "a")
		
		writeLogFile(logfile, "\n\n************ LOG ************\n\n", "a")
		writeTime(start, logfile)
		writeLogFile(logfile, "Start\n", "a")

		sys.stdout.write("\nPre-matching Started........")
		sys.stdout.flush()

		pre_start=time()
		writeTime(pre_start, logfile)
		writeLogFile(logfile, "Pre-matching started\n", "a")
		preExact(rows1,rows2, matchcol1, matchcol2, n_threads, tmp_path)
		phase1_time=time()-start
		print(" DONE (duration: %4.2fs)\n" % (phase1_time))
		writeLogFile(logfile, "\t\tPre-matching phase duration: %4.2fs" % (phase1_time)+"\n", "a")

		sys.stdout.write("\nCatalogues Matching Started...")
		sys.stdout.flush()
		t1=time()
		writeTime(t1, logfile)
		writeLogFile(logfile, "Crossmatching started\n", "a")

		exactRun(n_threads, tmp_path)

		t_match=time()-t1
		print(" DONE (duration: %4.2fs)\n" % (t_match))
		writeLogFile(logfile, "\t\tCrossmatching phase duration: %4.2fs" % (t_match)+"\n", "a")

		des1=-99
		des2=-99
		outdata=[selection, join, header1, header2, des1, des2]

		n_matches=createOutput(tmp_path, stiltsDir, outputFile, outformat, n_threads, logfile, rows1,rows2, criteria, outdata)


		### Removing temporary files
		removeTmp(logfile, tmp_path)


	elif criteria=="sky":
		
		errorType, shape=readSkyType(configfile, logfile)
		selection, join=readSelection(configfile, logfile)
		cellwidth=getCellSize(configfile, logfile, tmp_path)

		print("Error type: "+errorType+" - Area shape: "+shape)
		print("Match Selection: "+selection+ " - Join Type: "+join)
		print("Minimum Partition Cell Size: " + str(cellwidth) +" arcsecs")

		
		writeLogFile(logfile, "Error type: "+errorType+"\n", "a")
		writeLogFile(logfile, "Area shape: " + shape +"\n", "a") 
		writeLogFile(logfile, "Match Selection: " + selection +"\n", "a")
		writeLogFile(logfile, "Join Type: " + join +"\n", "a") 	
		writeLogFile(logfile, "Minimum Partition Cell Size: " + str(cellwidth) +" arcsecs\n", "a") 


		writeLogFile(logfile, "\n\n************ LOG ************\n\n", "a")
		writeTime(start, logfile)
		writeLogFile(logfile, "Start\n", "a")

		sys.stdout.write("\nPre-matching Started........")
		sys.stdout.flush()

		pre_start=time()
		writeTime(pre_start, logfile)
		writeLogFile(logfile, "Pre-matching started\n", "a")


		if errorType=="fixed":
			axis1, axis2 = readFixedValues(configfile, logfile)
			fwhma_col=axis1
			fwhmb_col=axis2
			fwhma_min=axis1
			fwhma_max=axis1
			fwhmb_min=axis2
			fwhmb_max=axis2
			fwhm_max=max(axis1,axis2)
			factor=1.0
		else:
			fwhma_col, fwhmb_col, factor= readFWHMValues(header1, configfile, logfile, tmp_path)
			fwhma_min, fwhma_max=findColMinMax(rows1, fwhma_col)
			fwhmb_min, fwhmb_max=findColMinMax(rows1, fwhmb_col)
			fwhm_max=max(fwhma_max, fwhmb_max) 			



		if fwhm_max>cellwidth:
			width=fwhm_max			
		else:
			width=cellwidth
		
		maxAreaSize=factor*width

		coordsys1, glon1_col, glat1_col, units1=readCoordsPars(configfile, logfile, 1, header1,tmp_path)
		coordsys2, glon2_col, glat2_col, units2=readCoordsPars(configfile, logfile, 2, header2,tmp_path)
		des1, des2=getDesignationColumn(header1, header2, configfile, logfile,tmp_path)
		pa_col, clock, shift=readPA(configfile, logfile, header1, errorType,tmp_path)

		newRowsCat1(rows1, file1tmp, coordsys1, units1, des1, glon1_col, glat1_col, fwhma_col, fwhmb_col, errorType, factor, pa_col, tmp_path, header1, stiltsDir, logfile, n_threads)

		pythonexec=sys.executable
		cmd=pythonexec+" prematch.py "+file2tmp+" "+coordsys2+" "+units2+" "+str(des2)+" "+str(glon2_col)+" "+str(glat2_col)+" "+tmp_path+" "+stiltsDir+" "+logfile+" "+str(maxAreaSize)+" "+str(len(rows2))

		system(cmd)



		phase1_time=time()-start
		print(" DONE (duration: %4.2fs)\n" % (phase1_time))
		writeLogFile(logfile, "\t\tPre-matching phase duration: %4.2fs" % (phase1_time)+"\n", "a")


		sys.stdout.write("Catalogues Matching Started...")
		sys.stdout.flush()
		t1=time()
		writeTime(t1, logfile)
		writeLogFile(logfile, "Crossmatching started\n", "a")




		area_params=[shape, clock, shift]

		outdata=[selection, join, header1, header2, des1, des2]

		skyRun(area_params, tmp_path,n_threads, maindir, outdata)  #Main threaded process
		t_match=time()-t1
		print(" DONE (duration: %4.2fs)\n" % (t_match))
		writeLogFile(logfile, "\t\tCrossmatching phase duration: %4.2fs" % (t_match)+"\n", "a")


		sys.stdout.write("Creating Output........")
		sys.stdout.flush()
		t2=time()
		writeTime(t2, logfile)
		writeLogFile(logfile, "Output creation started\n", "a")
		
		n_matches=createOutput(tmp_path, stiltsDir, outputFile, outformat, n_threads, logfile, rows1,rows2, criteria, outdata)
		
		t_out=time()-t2
		print(" DONE (duration: %4.2fs)\n" % (t_out))
		writeLogFile(logfile, "\t\tOutput creation phase duration: %4.2fs" % (t_out)+"\n", "a")

		### Removing temporary files
		removeTmp(logfile, tmp_path)


	print("Number of Output Rows: "+str(n_matches)+"\n")
	writeLogFile(logfile, "Number of Output Rows: "+str(n_matches)+"\n\n","a")

	print("\nCOMPLETED!")
	
	tempo=time()-start
	writeLogFile(logfile, "C3 End Time: "+asctime(localtime(time()) )+"\n\n", "a")
	writeLogFile(logfile, "Elapsed time: %4.2fs\n" % (tempo), "a")

	print("\nTotal elapsed time: %4.2f\n" % (tempo))


if __name__ == "__main__":
	parser = argparse.ArgumentParser(prog='python c3.py')

	parser.add_argument('-c', '--config', metavar="FILE", help='Configuration file name')
	parser.add_argument('-d', '--default', metavar="FILE", nargs='?', const='config.ini', help='To create default configuration file')
	
	args = parser.parse_args()
	
	if len(sys.argv)>3:
		print("\nError! It is possible to use only 1 argument at a time\n")
		parser.print_help()
		exit(1)

	maindir=sep.join(sys.argv[0].split(sep)[:-1])+sep
	if maindir==sep:
		maindir="."+sep
	configfile = vars(args)['config']
	dummy=vars(args)['default']
	if dummy!=None:
		createConfig(dummy)
	else:
		main(configfile, maindir)
